"""
Tiferet.py
An exercise in writing formulaic stories according to a certain stereotype.
Basically what this does is form a "dictionary" describing the TFd character's
starting and ending forms. The dictionaries (also called hash tables) describe
a set of body parts, each with a color and other attributes. The main part
of the program, "Tiferet," runs a series of turns where the character has one
attribute of one part change each time. Each change is built into an English
sentence using some ad-hoc sentence formats.
In this program there's no plot or AI whatsoever.

Version 2: Incorporates XML character-building code and (unused) decision-making
code from Chispa v2, so that characters' descriptions are built from templates
for greater variety. This version also includes a non-obvious fix from an
earlier version of the character-builder, involving default body coverings
being applied to body parts before said defaults were finished being built.
"""

## INCLUDED FILES ##
## Standard Python modules.
import string
import copy
from random import randint

from ab_xml import *



## CONSTANTS ##

TEMPLATE_DIRECTORY = "templates"

FORM1 = {"species":"human"}
FORM2 = {"species":"ferret"}

PRONOUNS_NOM = {"male":"he","female":"she","herm":"shi","neuter":"it"}
PRONOUNS_ACC = {"male":"his","female":"her","herm":"hir","neuter":"its"}
SIZES = ["tiny","small","smallish","large","huge"]
PLURAL = ["legs","arms","hands","feet","eyes","ears","fingers","toes","nails","claws","nostrils","hips","shoulders"]
INDENT = "    "

## An experiment with greater variety of sentence structure. Didn't work
## properly b/c it was inflexible with plurals etc.
SENTENCE_OUTLINES = [("$subject did stuff.",[]),
                     ("$subject didn't do stuff.",["neg"]),
                     ("$subject's $part turned $adj.",["tf"]),
                     ("$subject saw a $part appear.",["tf","appear"]),
                     ("$subject saw their $part vanish.",["tf","vanish"]),
                     
                     ]

DEFAULT_INTERESTS = {"person":2,"dangerous":-9,"food":5,"shiny":3,"cute":3,"interesting":1}
DEFAULT_BELIEFS = {"nature:person":"person"}


class Spark:
    def __init__(self,**options):
        """An intermediary experiment between Tiferet Lv.1 and Chispa.
        A description of a physical body, used by Tiferet, using info
        built from XML templates. The class also includes "beliefs," which
        are not used here."""
        self.name = options.get("name")
        self.interests = options.get("interests",DEFAULT_INTERESTS)
        self.beliefs = options.get("beliefs",DEFAULT_BELIEFS)
        self.messages = []
        self.past_messages = []

        self.data = options.get("data",{}) ## Supplied by CharBuilder.

    def GetName(self):
        return self.name
    def GetSex(self):
        return self.data["desc"]["sex"]
    def GetBreed(self):
        return self.data["desc"]["breed"]

    def Message(self,what):
        self.messages.append(what)
    def ClearMessages(self):
        self.messages = []
        self.past_messages = []
    def CycleMessages(self):
        self.past_messages = self.messages
        self.messages = []
    def ShowMessages(self):
        for message in self.messages:
            print message
    def DescribeMessage(self,message={}):
        """Assign tags to this message based on my impressions."""
        print message
        tags = []
        for key in message.keys():
##            print "Key: "+key
            ## Eg. "color"
            value = message[key]
##            print "Value: "+value
            if type(value) is str:
                ## If only one entry, treat as a one-item list.
                value = [value]
            for n in value:
                ## Eg. "blue":
                if key == "name":
                    search_term = n
                else:
                    search_term = key+":"+n
                print "Searching for: "+search_term
                belief = self.beliefs.get( search_term )
                if not belief:
                    pass
                elif type(belief) is str:
                    print "Belief: "+str([belief])
                    tags += [belief]
                else:
                    print "Belief: "+str(belief)
                    tags += belief
        print "Tags: "+str(tags)
        message["tags"] = tags
        return message
    def Go(self):
        """Select a message to focus on."""
        high_score = -9999
        highest_appeal = -9999
        focus = None
        for message in self.messages:
##            print "\nDescribing message: "+str(message)
            message = self.DescribeMessage( message )
##            print "Judging message."
            positive_score = 0
            neutral_score = 0 #?
            negative_score = 0
            tags = message.get("tags")
            for tag in tags:
##                print tag
                multiplier = 1
                if tag and (tag[0] == "+"):
                    tag = tag[1:]
                    multiplier = 2
                elif tag and (tag[0] == "?"):
                    tag = tag[1:]
                    multiplier = .5
                elif tag and (tag[0] == "!"):
                    tag = tag[1:]
                    multiplier = -1

                my_interest = self.interests.get(tag)
                if my_interest != None:

                    my_interest = int(my_interest * multiplier)
                    print my_interest

                    if my_interest > 0:
                        positive_score += my_interest
                    elif my_interest < 0:
                        negative_score -= my_interest
                    else:
                        neutral_interest += 1

            interest = positive_score + abs(negative_score) + neutral_score
            appeal = positive_score - negative_score
##            print "All scores: "+str(positive_score)+"/"+str(neutral_score)+"/"+str(negative_score)
##            print "Message rating: Interest: "+str(interest)+", Appeal: "+str(appeal)

            if (interest > high_score) or ((interest == high_score) and (appeal > highest_appeal) ):
                ## Copy this message so we can alter it freely.
                focus = copy.copy(message)
                high_score = interest
                highest_appeal = appeal

                ## Rate the target by how interesting it is, and how much I like it.
                focus["interest"] = interest
                focus["appeal"] = appeal

        print "Focusing on message "+str(focus)+" (interest: "+str(interest)+", appeal: "+str(appeal)+")"

        self.focus = focus

    def Act(self):
        if not self.focus:
            print "Error: No focus."
            return
        interest = self.focus.get("interest")
        appeal = self.focus.get("appeal")
        if (interest == None) or (appeal == None):
            print "Error: Blank focus?"
            return

        if appeal > 0:
            print "I like it!"
        elif appeal == 0:
            print "Eh, I'm ambivalent."
        else:
            print "I don't like it!"        




class CharBuilder:
    def __init__(self,**options):
        """Creates a description of a character's beliefs and body,
        based on XML templates."""
        self.templates = {}
        if options.get("load_templates",True):
            self.LoadCharTemplates()

    def LoadCharTemplates(self):
        file_list = os.listdir( os.path.join(os.getcwd(),TEMPLATE_DIRECTORY) )
        for filename in file_list:
            if filename[-3:].lower() != "xml":
                continue
            print "*** Loading file: "+filename+" ***"
            xml_file = file( os.path.join(TEMPLATE_DIRECTORY, filename) , "r" )
            templates_rawdata = GrabAllObjects(xml_file,"template")

            for data in templates_rawdata:
                traits, trait_names = GrabObjectTraits(data)
                if traits:
                    content = dict([(trait_names[n],traits[n]) for n in range(len(traits))])

                    template = {}

                    if content.has_key("name"):
                        template["name"] = content.get("name")
                    else:
                        template["name"] = "Unknown Template "+str(randint(0,999))

                    if content.has_key("nature"):
                        template["nature"] = content.get("nature")
                    else:
                        template["nature"] = ""

                    if content.has_key("beliefs"):
                        beliefs_data = eval(content["beliefs"])
                        beliefs = {}
                        for b in beliefs_data:
                            beliefs[ b ] = beliefs_data[ b ]
                        template["beliefs"] = beliefs

                    if content.has_key("parts"):
                        parts_data = eval(content["parts"])
                        parts = {}
                        for p in parts_data:
                            parts[ p ] = parts_data[ p ]
                        template["parts"] = parts

                    if content.has_key("covering"):
                        template["covering"] = eval(content["covering"])

##                    ## Debug: Report.
##                    print "\n--- Loaded template: "+template["name"]+" ("+template["nature"]+") ---"
##                    for key in template:
##                        if key not in ["name","nature"]:
##                            value = template[key]
##                            if type(value) in (str,int):
##                                print "-"+key+": "+str(value)
##                            else:
##                                print key+":"
##                                for innerkey in value:
##                                    print "    -"+innerkey+": "+str(value[innerkey])

                    ## We've extracted what info we can. Store this breed.
                    self.templates[ template["name"] ] = template

            xml_file.close()

    def Build(self,template_list=[],**options):
        """Create a character by overlaying templates.
        By inputting something like "male human", we get a description of
        the character's body and attitudes -- details like the color of
        their hair or whether they think of meat as food.

        If an attribute is given as a list, need to pick a random item from it.
        Eg., for hair color.

        Beliefs should be merged into a single beliefs set for the character,
        but the character should also keep separate copies so that we can
        notice interesting internal disputes: "As a hunter he knew the carcass
        to be food, but as a squirrel he found the idea revolting."

        Body parts should be merged into a single set.
        Where a part already exists, sizes are added (w/limits).
        """

        name = options.get("name","Grignr")
        data_only = options.get("data_only")

        ## Handle string input like "male fox".
        if type(template_list) is str:
            template_list = template_list.strip(",").split()

        print "Building character "+name+" "+str(template_list)+"."

        ## Always apply the default template first.
        template_list = ["default"] + template_list

        ## Begin building the character.
        char = {}
        desc = {}
        beliefs = {}
        parts = {}

        body_covering = {}

        ## For each template...
        for t in template_list:
            template = copy.deepcopy( self.templates.get( t ) )
            if not template:
                print "Template '"+t+"' not found."
                continue

            nature = template["nature"]
            desc[ nature ] = t ## Eg., stores that "breed" is "fox".

            print "Applying template: "+t+" (Nature = "+str(nature)+")"

            ## Handle beliefs.
            template_beliefs = template.get("beliefs")
            if template_beliefs:
                for key in template_beliefs:
                    current_value = beliefs.get(key,[])
                    this_template_value = template_beliefs[ key ]
                    if type(this_template_value) is str:
                        this_template_value = [this_template_value]
                    beliefs[ key ] = current_value + this_template_value

            ## Handle body covering, a set of defaults.
            ## The default covering for each body part is built via multiple
            ## templates, then applied after all templates are loaded.
            covering = template.get("covering",{})
            body_covering.update(covering)
##            print "Updating covering data to: "+str(body_covering)

            ## Handle parts. (Pun not intended.)
            ## Debug: Watch carefully to make sure old templates aren't overwritten.
            ## Bug: Numeric values aren't being summed yet.
            template_parts = template.get("parts")
            if template_parts:
                for key in template_parts:
                    current_part_data = copy.deepcopy(parts.get(key))
                    this_template_data = template_parts[ key ]
##                    print "Re: part: "+key
##                    print "Existing part data: "+str(current_part_data)
##                    print "This template data: "+str(this_template_data)
                    if current_part_data:
                        ## This part already exists.
##                        new_part_data = this_template_data ## Temp.
                        ## "update" is in-place, so we'll sort of copy c_p_d.
                        current_part_data.update( this_template_data )
                        new_part_data = current_part_data
                    else:
                        new_part_data = this_template_data
##                    print "New part data: "+str(new_part_data)

                    ## Replace list entries by picking a value.
                    ## Eg., pick an eye color from ["blue","grey" etc.].
                    for part_detail in new_part_data:
                        value = new_part_data[ part_detail ]
                        if type( value ) is list:
                            random_choice = value[ randint(0,len(value)-1) ]
                            new_part_data[ part_detail ] = random_choice
                            print "Chose '"+random_choice+"' for "+key+"."+part_detail+"."
                    parts[ key ] = new_part_data

        ## We've now applied all templates, so we know what parts exist and
        ## what they look like by default. Apply that data.
        ## Eg., each part of a fox is some fur color by default.

        ## First, specifify values if any are in list form.
##        print "Applying body covering: "+str(body_covering)
        for key in body_covering:
            value = body_covering[key]
            if type( value ) is list:
                random_choice = value[ randint(0,len(value)-1) ]
                body_covering[ key ] = random_choice
##                print "Specified that '"+key+"' is "+random_choice+"."
        for key in parts:
            part = parts[key]
            for default_trait in body_covering:
                if not part.has_key( default_trait ):
                    ## Put in the default.
                    parts[key][default_trait] = body_covering[default_trait]


        data = {"desc":desc,"beliefs":beliefs,"parts":parts}
        if data_only:
            return data
        char = Spark(name=name,data=data)
        return char


class Tiferet:
    def __init__(self,**options):
        """Tiferet: Full-time TF Ferretaur!"""
        self.name = options.get("name","Tiferet")
        self.title = options.get("title","Full-Time TF Ferretaur")
        self.sequence = []

    def MyName(self,full=False):
        name = self.name
        if full and self.title:
            name += ", " + self.title
        return name

    def ShowOff(self):
        """Random bit of text."""
        choices = ["takes a bow","smiles","grins","nods sagely"]
        n = randint(0,len(choices)-1)
        return choices[n]

    def Subject(self,char,poss=False):
        """Grammatical widget."""
        n = randint(0,1)
        if n == 0:
            if poss:
                return char.GetName()+"'s"
            else:
                return char.GetName()
        else:
            if poss:
                return PRONOUNS_ACC[ char.GetSex() ]
            else:
                return PRONOUNS_NOM[ char.GetSex() ]

    #### Three functions for sentence outlines; not currently used. ####

    def BuildSentenceFromOutline(self,outlines,**data):
        ## Pick an outline.
        if not outlines:
            outlines = ["I forgot to supply outlines."]
##        text = outlines[ randint(0,len(outlines)-1) ]
##        words = text.split()
##        endpunc = "."
##        for n in range(len(words)):
##            if (n == len(words)-1) and (words[n][-1] in ".?!"):
##                endpunc = words[n][-1]
##                words[n] = words[n][:-1]
##            print words[n]
##            if words[n][0] == "$":
##                value = data.get( words[n][1:] )
##                if value:
##                    words[n] = value
##                    
##        text = string.join(words)+endpunc

        text = outlines[ randint(0,len(outlines)-1) ]
        words = text.split()
        endpunc = "."
        for n in range(len(words)):
            print "Word (unmodified): "+words[n]
            ## Modify some words to handle punctuation.
            if words[n][-2:] == "'s":
                endpunc = "'s"
                words[n] = words[n][:-2]
            elif words[n][-1] in ".?!":
                endpunc = words[n][-1]
                words[n] = words[n][:-1]
            print "Word as modified via punctuation: "+words[n]
            if words[n][0] == "$":
                value = data.get( words[n][1:] )
                if value:
                    words[n] = value
                    
        text = string.join(words)
        if text[-1] not in ".?!":
            text += "."

        return text

    def PickSentenceOutline(self,tags=[]):
        """Select a sentence outline having this description."""
        usable_outlines = []
        for outline in SENTENCE_OUTLINES:
            OK = True
            print outline
            for tag in tags:
                if tag not in outline[1]:
                    OK = False
            if OK:
                usable_outlines.append( outline[0] )
        return usable_outlines

    def ComposeLine(self,tags=[],**data):
        return self.BuildSentenceFromOutline( self.PickSentenceOutline(tags), **data )

    ##################################################

    def WriteLine(self,line,special=None):
        """Store this line of text with some special options."""
        if special == "header":
            line = "--- "+line+" ---"
            self.sequence.append(line)
            self.sequence.append("")
        elif special == "endpara":
            self.sequence.append(line)
            self.sequence.append("")
        elif special == "newpara":
            self.sequence.append("")
            self.sequence.append(line)
        else:
            self.sequence.append(line)

    def TF(self,char,newform):
        """Write a scene describing char changing into newform.
        Note that this doesn't directly change char, the data structure."""
        shape1 = char.data["parts"]
        shape2 = newform.data["parts"]

        self.WriteLine(char.GetName()+" Becomes a "+newform.GetBreed().title(),"header")
        self.WriteLine("by "+self.MyName(True),"header")
        self.WriteLine(char.GetName()+" is a "+char.GetBreed()+".")

        done = False
        if shape1 == shape2:
            self.WriteLine(self.MyName()+" blinks. \"Uh, I don't really have to do anything.\"")
            done = True

        else:
            self.WriteLine(self.MyName()+" zaps "+char.GetName()+" and...")

        sequence = []

        rounds = 0
        while not done:
##            print "--- Round "+str(rounds)+" ---"
            rounds += 1

##            if rounds > 30:
##                done = True

            ## A precise description of the change.
            result = {"subject":char,"new_value":[]}

            targets = []
            for key in shape1.keys() + shape2.keys():
                if shape1.get(key) != shape2.get(key):
                    if key not in targets:
                        targets.append(key)
##            print "Targets: " + str(targets)
            if not targets:
##                print "Done."
                done = True
            else:
                n = randint(0,len(targets)-1)
                target = targets[n]
##                print "Selected: " + target
                result["changed_part"] = target
                part_before = shape1.get(target)
                part_after = shape2.get(target)
##                print "Part: "+target+" Before: "+str(part_before)+"\nAfter: "+str(part_after)
                if not part_before:
                    ## This part doesn't exist yet! Create it.
                    result["change_type"] = "appear"
                    shape1[target] = copy.deepcopy(shape2[target])
                    shape1[target]["size"] = 0
                elif not part_after:
                    ## This part vanishes!
                    result["change_type"] = "vanish"
                    del shape1[target]
                else:
                    ## Consider doing several changes at once.
                    max_loops = [1,1,2,3][ randint(0,3) ]
                    loops_so_far = 0
##                    print "Changing part: "+target
##                    print "Max loops: "+str(max_loops)

                    while loops_so_far < max_loops:
                        ## Select a trait to change.
                        ## Valid traits have a new and different value.
                        part_before = shape1.get(target)
                        part_after = shape2.get(target)
##                        print "Part: "+target+" Before: "+str(part_before)+"\nAfter: "+str(part_after)

                        traits = []
                        for key in part_before:
                            if (part_after.has_key(key)) and (part_before.get(key) != part_after.get(key)):
                                traits.append(key)
                        for key in part_after:
                            if (not key in traits) and (part_before.get(key) != part_after.get(key)):
                                traits.append(key)
                        if not traits:
                            """Most likely because the new form is missing some
                            defined trait of the old, the part registers as
                            different but effectively isn't. Make it ==, so that
                            it won't keep getting selected for change."""
##                            print "Parts set equal."
                            shape1[target] = shape2[target]
                        else:
##                            print "Choosing from different traits: "+str(traits)
                            n = randint(0,len(traits)-1)
                            trait = traits[n]
##                            print "Selected trait to change: " + trait
                            new_value = shape2[target][trait]
                            shape1[target][trait] = new_value
                            result["change_type"] = "shift"
                            result["new_value"].append(new_value)
                        loops_so_far += 1

                sequence.append( result )

        print "Done with composition. Now the writing."
##        print "Sequence: "+str(sequence)
        for step in sequence:
##            print step
            text = self.ConvertSequenceLineToText( step )
            if text:
                self.WriteLine( text )

        self.WriteLine(self.MyName()+" "+self.ShowOff()+", declaring "+char.GetName()+" a "+newform.GetBreed()+"!")
        self.WriteLine("The End.")

        self.Write()

    def ConvertSequenceLineToText(self,step={}):
        """In progress: take a description of a change step
        and return a sentence."""

        char = step.get("subject")
        change_type = step.get("change_type")
        changed_part = step.get("changed_part")
        new_value = step.get("new_value")

        if not change_type:
            return ""

        elif change_type == "appear":
            subject = self.Subject(char)
            text = subject+" feels a "+changed_part+" appear."
##            text = self.ComposeLine( ["tf","appear"], subject=subject,part=target )
##            print "Text: "+text
        elif change_type == "vanish":
            ## This part vanishes!
            subject = self.Subject(char)
            text = subject+" feels their "+changed_part+" vanish."
##            text = self.ComposeLine( ["tf","vanish"], subject=subject,part=target )
        else:
            for n in range(len(new_value)):
                if type(new_value[n]) is int:
                    ## Turn numeric size into an adjective.
                    new_value[n] = SIZES[new_value[n]]
            new_value_text = FancyList(new_value)

            if changed_part in PLURAL:
                verb = " turn "
            else:
                verb = " turns "
            subject = self.Subject(char,True)
            text = subject+" "+changed_part+verb+new_value_text+"."
##            text = self.ComposeLine( ["tf"], subject=subject,part=target,adj=new_value_text )

        return text

    def Write(self):
        """Turn the cached lines into a finished string of text.
        This change includes stuff like random paragraph breaks."""
        result = INDENT
        for line in self.sequence:
##            print "Writing line: "+"'"+str(line)+"'"

            n = randint(0,3)
            if n == 0:
                result += "\n"+INDENT

            if line:
                line = line[0].upper()+line[1:]
                result += line+" "
            else:
                result += "\n"+INDENT

        print result


## AUTORUN CODE ##

## Set up a demo.
tiferet = Tiferet(form1={"species":"human"},
                  form2={"species":"ferret"},
          )

## Run demo.
##tiferet.Go()

tiferet.BuildSentenceFromOutline(tiferet.PickSentenceOutline(),**{"act":"write","name":"K"})

builder = CharBuilder()
start_form = builder.Build("human male",name="Grignr")
final_form = builder.Build("fox male")

tiferet = Tiferet()
tiferet.TF(start_form, final_form)

"""Stuff re: the Chispa v2 decision-making system, which isn't used here.
It doesn't really work in this context because the characters, here, have
only the default set of beliefs."""
##f = {"nature":"person","name":"Kay","species":"raccoon","sex":"female"}
##m = {"nature":"person","name":"Snow","species":"raccoon","sex":"male"}
##o = {"nature":"person","name":"Atax","species":"ogre","sex":"male"}
##k.Message(m)
##k.Message(f)
##k.Message(o)
##k.Go()
##k.Act()
